/* ------------------------------------------------------------------------- */
/*!
 *  \file       atlas_generator.cc
 *  \brief      アトラスジェネレータ
 *  \date       Since: June 7, 2021. 17:15:06 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include "atlas_generator.h"

#include <iterator>
#include <json/json.h>
#include <mtk/byte_stream.h>
#include <mtk/file_pointer.h>
#include <mtk/text_hash.h>

#include "message.h"

//#define ENABLE_DEBUG_TRACE

#if defined(ENABLE_DEBUG_TRACE)
#define DEBUG_TRACE     printf
#else
#define DEBUG_TRACE(...)
#endif

namespace texatlas
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      ターゲットリストを構築
 *  \param[in]  targetFiles     対象ファイル
 *  \param[in]  hashSeed        ハッシュシード値
 *  \return     対象となるファイルの数
 */
/* ------------------------------------------------------------------------- */
size_t AtlasGenerator::CreateTargetList(const std::vector<std::filesystem::path> &targetFiles, uint32_t hashSeed) noexcept
{
    // 画像ファイルの情報を構築してリストに追加
    for (const auto &target : targetFiles)
    {
        ImageFileInfo info(target, hashSeed);

        if (!info)
        {
            PutError(MsgID::FailedToReadFile, {target});
            return 0;
        }

        _targetList.push_back(info);
    }

    // サイズを元にリストをソート
    // 高さが大きいものを優先し，高さが同じであれば幅が大きい方を優先する
    auto sortFunction = [](const ImageFileInfo &a, const ImageFileInfo &b)
    {
        bool swap = false;

        if (a.GetHeight() == b.GetHeight())
        {
            swap = a.GetWidth() > b.GetWidth();
        }
        else
        {
            swap = a.GetHeight() > b.GetHeight();
        }

        return swap;
    };
    _targetList.sort(sortFunction);

    // ハッシュシード値はファイルに出力するので保持しておく
    _hashSeed = hashSeed;

    return _targetList.size();
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      ターゲットリストを表示
 */
/* ------------------------------------------------------------------------- */
void AtlasGenerator::PrintTargetList() const noexcept
{
    for (const auto &target : _targetList)
    {
        target.Print();
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      ハッシュの衝突チェック
 *  \return     検出数
 */
/* ------------------------------------------------------------------------- */
size_t AtlasGenerator::CheckCollidingHash() const noexcept
{
    size_t count = 0;

    for (auto target1 = _targetList.begin(); target1 != _targetList.end(); ++target1)
    {
        for (auto target2 = std::next(target1); target2 != _targetList.end(); ++target2)
        {
            if (target1->GetHash() == target2->GetHash())
            {
                PutError(MsgID::HashCollisionDetected, {target1->GetName(), target2->GetName(), target1->GetHash()});
                count++;
            }

            for (const auto &child : target2->GetChildren())
            {
                if (target1->GetHash() == child.hash)
                {
                    PutError(MsgID::HashCollisionDetected, {target1->GetName(), child.name, target1->GetHash()});
                    count++;
                }
            }
        }
    }

    return count;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      アトラス化の実行
 *  \param[in]  width       生成するテクスチャの幅
 *  \param[in]  height      生成するテクスチャの高さ
 *  \param[in]  margin      マージン
 *  \param[in]  outputName  出力名
 *  \retval     true        成功
 *  \retval     false       失敗
 */
/* ------------------------------------------------------------------------- */
bool AtlasGenerator::Generate(uint32_t width, uint32_t height, uint32_t margin, const char *outputName) noexcept
{
    // サイズに収まらない画像が無いかチェック
    bool existOverImage = false;
    for (const auto &target : _targetList)
    {
        if (((target.GetWidth() + margin) >= width) ||
            ((target.GetHeight() + margin) >= height))
        {
            PutError(MsgID::NotFitImage, {target.GetPath(), target.GetWidth(), target.GetHeight()});
            existOverImage = true;
        }
    }
    if (existOverImage)
    {
        return false;
    }

    // 矩形リストとビットマップを初期化
    RectangleList rectangles;
    rectangles.push_back({margin, margin, width, height});
    Bitmap bitmap(width, height, Bitmap::PixelFormat::RGBA);
    uint32_t textureIndex = 0;

    // 対象ファイルが無くなるまで繰り返す
    while (!_targetList.empty())
    {
        auto target = _targetList.begin();

        if (_isVerbose)
        {
            PutMessage(MsgID::PlaceIn, {outputName, textureIndex, target->GetPath()});
        }

        // 対象ファイルを読み込み
        Bitmap targetImage;
        targetImage.LoadPNG(target->GetPath());
        if (!targetImage)
        {
            PutError(MsgID::FailedToReadFile, {target->GetPath()});
            _targetList.pop_front();
            continue;
        }

        // 奇数サイズの画像に対して警告を出す
        if (((targetImage.GetWidth() & 1) != 0) || ((targetImage.GetHeight() & 1) != 0))
        {
            PutWarning(MsgID::NotEvenNumberSize, {target->GetPath(), targetImage.GetWidth(), targetImage.GetHeight()});
        }

        // 対象ファイルを収める丁度良いサイズの矩形を取得
        Rectangle allocRect(0, 0, target->GetWidth() + margin, target->GetHeight() + margin);
        auto fitRect = rectangles.PopFitRectangle(allocRect);
        if (!fitRect)
        {
            // 丁度良い矩形が無かったら新しい画像と矩形を生成
            _bitmapArray.push_back(std::move(bitmap));
            bitmap = Bitmap(width, height, Bitmap::PixelFormat::RGBA);
            rectangles.clear();
            rectangles.push_back({margin, margin, width, height});
            fitRect = rectangles.PopFitRectangle(allocRect);
            textureIndex++;

            DEBUG_TRACE("new texture\n");
        }

        // 画像をコピー
        Rectangle imageRect(0, 0, target->GetWidth(), target->GetHeight());
        DEBUG_TRACE("target size: %d, %d\n", imageRect.width, imageRect.height);
        DEBUG_TRACE("alloc size: %d, %d (%llX)\n", allocRect.width, allocRect.height, allocRect.GetSortKey());
        DEBUG_TRACE("fit rectangle: %d, %d, %d, %d (%llX)\n", fitRect.x, fitRect.y, fitRect.width, fitRect.height, fitRect.GetSortKey());
        bitmap.Copy(fitRect.x, fitRect.y, targetImage, imageRect.x, imageRect.y, imageRect.width, imageRect.height);

        // 配置情報を保存
        Location location;
        location.name = target->GetName();
        location.nameHash = target->GetHash();
        location.index = textureIndex;
        location.x = fitRect.x;
        location.y = fitRect.y;
        location.width = imageRect.width;
        location.height = imageRect.height;
        _locations.push_back(location);

        // 対象が子の配置情報を持っている場合はそれも保存
        for (const auto &child : target->GetChildren())
        {
            Location childLocation;
            childLocation.name = child.name;
            childLocation.nameHash = child.hash;
            childLocation.index = textureIndex;
            childLocation.x = child.x + fitRect.x;
            childLocation.y = child.y + fitRect.y;
            childLocation.width = child.width;
            childLocation.height = child.height;
            _locations.push_back(childLocation);

            // こちらも奇数サイズに対して警告を出す
            if (((child.width & 1) != 0) || ((child.height & 1) != 0))
            {
                PutWarning(MsgID::NotEvenNumberSize, {child.name, child.width, child.height});
            }
        }

        // 矩形を分割
        auto splitRects = fitRect.Split(allocRect);
        for (auto splitRect : splitRects)
        {
            rectangles.Add(splitRect);
        }
#if defined(ENABLE_DEBUG_TRACE)
        int count = 0;
        for (auto rectangle : rectangles)
        {
            DEBUG_TRACE("%2d: %4d, %4d, %4d, %4d (%16llX)\n", count++, rectangle.x, rectangle.y, rectangle.width, rectangle.height, rectangle.GetSortKey());
        }
#endif

        // 処理が完了したのでリストから削除
        _targetList.pop_front();

        DEBUG_TRACE("\n");
    }

    // 全ての処理が完了したら保存するビットマップに登録
    _bitmapArray.push_back(std::move(bitmap));

    // 配置情報は使う際に検索しやすいようハッシュ値を元にソートしておく
    _locations.sort([](const Location &a, const Location &b)
    {
        return a.nameHash < b.nameHash;
    });

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      保存
 *  \param[in]  outputPath  出力ディレクトリ
 *  \param[in]  name        出力名
 *  \retval     true        成功
 *  \retval     false       失敗
 */
/* ------------------------------------------------------------------------- */
bool AtlasGenerator::Save(const std::filesystem::path &outputPath, const std::filesystem::path &name) noexcept
{
    // 画像を出力
    for (size_t i = 0; i < _bitmapArray.size(); i++)
    {
        auto path = outputPath;
        path /= name;
        path += std::to_string(i) + ".png";

        if (!_bitmapArray[i].SavePNG(path))
        {
            PutError(MsgID::FailedToOutput, {path});
            return false;
        }

        PutMessage(MsgID::OutputFile, {path});
    }

    // 配置情報をJSONで出力
    if (!SaveLocatonToJson(outputPath, name))
    {
        return false;
    }

    // 配置情報をバイナリで出力
    if (!SaveLocatonToBinary(outputPath, name))
    {
        return false;
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      配置情報をJSONで保存
 *  \param[in]  outputPath  出力ディレクトリ
 *  \param[in]  name        出力名
 *  \retval     true        成功
 *  \retval     false       失敗
 */
/* ------------------------------------------------------------------------- */
bool AtlasGenerator::SaveLocatonToJson(const std::filesystem::path &outputPath, const std::filesystem::path &name) const noexcept
{
    Json::Value rootValue;

    auto &hashInfo = rootValue["hash"];
    hashInfo["algorithm"] = "FNV1a32";
    hashInfo["seed"] = _hashSeed;
    hashInfo["prime"] = MTK::Hash::kFNV1aPrime32;

    int i = 0;
    for (const auto &info : _locations)
    {
        auto &value = rootValue["locations"][i++];

        value["name"] = info.name;
        value["hash"] = info.nameHash;
        value["index"] = info.index;
        value["x"] = info.x;
        value["y"] = info.y;
        value["width"] = info.width;
        value["height"] = info.height;
    }

    // パスを生成
    auto path = outputPath;
    path /= name;
    path += "_loc.json";

    // 保存ファイルをオープン
    MTK::File::Pointer fp(path, "w");
    if (!fp)
    {
        PutError(MsgID::FailedToOutput, {path});
        return false;
    }

    // 書き込み
    Json::StreamWriterBuilder builder;
    auto saveString = Json::writeString(builder, rootValue);
    fwrite(saveString.c_str(), saveString.size(), 1, fp);

    PutMessage(MsgID::OutputFile, {path});

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      配置情報をバイナリで出力
 *  \param[in]  outputPath  出力ディレクトリ
 *  \param[in]  name        出力名
 *  \retval     true        成功
 *  \retval     false       失敗
 */
/* ------------------------------------------------------------------------- */
bool AtlasGenerator::SaveLocatonToBinary(const std::filesystem::path &outputPath, const std::filesystem::path &name) const noexcept
{
    constexpr size_t kBufferSize = 8 * 1024 * 1024;     // とりあえず多めに
    auto buffer = std::make_unique<std::byte[]>(kBufferSize);
    MTK::ByteStream stream(buffer.get(), kBufferSize);

    stream.Write(_hashSeed);
    stream.Write(MTK::Hash::FNV1a("TextureAtlasLocations", _hashSeed));
    stream.Write(uint32_t(0));  // リビジョン情報（予約）
    stream.Write(uint32_t(0));  // フラグ情報（予約）
    stream.Write(static_cast<uint32_t>(_bitmapArray.size()));
    stream.Write(static_cast<uint32_t>(_locations.size()));
    
    for (const auto &location : _locations)
    {
        stream.Write(location.nameHash);
        stream.Write(location.index);
        stream.Write(location.x);
        stream.Write(location.y);
        stream.Write(location.width);
        stream.Write(location.height);
        stream.Write(uint32_t(0));  // フラグ情報（予約）
    }

    stream.Write(MTK::Hash::FNV1a("EndOfRecord", _hashSeed));

    // パスを生成
    auto path = outputPath;
    path /= name;
    path += "_loc.bin";

    // ファイルをオープン
    MTK::File::Pointer fp(path, "wb");
    if (!fp)
    {
        PutError(MsgID::FailedToOutput, {path});
        return false;
    }

    fwrite(buffer.get(), stream.GetOffset(), 1, fp);

    PutMessage(MsgID::OutputFile, {path});

    return true;
}

}   // namespace

// vim: et ts=4 sw=4 sts=4
