/* ------------------------------------------------------------------------- */
/*!
 *  \file       bitmap.cc
 *  \brief      ビットマップクラス
 *  \date       Since: September 18, 2018. 22:45:23 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include "bitmap.h"

#include <string.h>

#include <png.h>
#include <mtk/file_pointer.h>
#include <mtk/text_format.h>

#include "message.h"

namespace texatlas
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      コンストラクタ
 *  \param[in]	width		幅
 *  \param[in]	height		高さ
 *  \param[in]	pixelFormat	ピクセルフォーマット
 */
/* ------------------------------------------------------------------------- */
Bitmap::Bitmap(uint32_t width, uint32_t height, PixelFormat pixelFormat) noexcept
    : _isValid(false)
    , _pixelFormat(PixelFormat::Gray)
    , _pixelSize(0)
    , _pixelBuffer(nullptr)
    , _pixelBufferSize(0)
    , _width(0)
    , _height(0)
{
    Initialize(width, height, pixelFormat);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      初期化
 *  \param[in]	width		幅
 *  \param[in]	height		高さ
 *  \param[in]	pixelFormat	ピクセルフォーマット
 *  \retval		true		成功
 *  \retval		false		失敗
 */
/* ------------------------------------------------------------------------- */
bool Bitmap::Initialize(uint32_t width, uint32_t height, PixelFormat pixelFormat) noexcept
{
    if (_isValid)
    {
        return false;
    }

    switch (pixelFormat)
    {
        case PixelFormat::Gray:
            _pixelSize = 1;
            break;

        case PixelFormat::GrayAlpha:
            _pixelSize = 2;
            break;

        case PixelFormat::RGB:
            _pixelSize = 3;
            break;

        case PixelFormat::RGBA:
            _pixelSize = 4;
            break;

        default:
            return false;
    }

    _pixelFormat = pixelFormat;
    _width = width;
    _height = height;
    _pixelBufferSize = static_cast<size_t>(width * height) * _pixelSize;
    _pixelBuffer = std::make_unique<uint8_t[]>(_pixelBufferSize);
    memset(_pixelBuffer.get(), 0, _pixelBufferSize);

    _isValid = true;

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      ピクセルを設定
 *  \param[in]	x		X座標
 *  \param[in]	y		Y座標
 *  \param[in]	red		赤
 *  \param[in]	green	緑
 *  \param[in]	blue	青
 *  \param[in]	alpha	アルファ（省略時は0xFF，無透過フォーマットの場合は無視）
 */
/* ------------------------------------------------------------------------- */
void Bitmap::SetPixel(uint32_t x, uint32_t y, const Color &color) noexcept
{
    if (!_isValid || !IsInCanvas(x, y))
    {
        return;
    }

    auto index = static_cast<size_t>(((y * _width) + x)) * _pixelSize;

    switch (_pixelFormat)
    {
        // グレースケール＋アルファ
        case PixelFormat::GrayAlpha:
            _pixelBuffer[index + 1] = color.alpha;
            [[fallthrough]];

        // グレースケール
        case PixelFormat::Gray:
            {
                _pixelBuffer[index] = color.ToGrayScale();
            }
            break;

        // RGBA
        case PixelFormat::RGBA:
            _pixelBuffer[index + 3] = color.alpha;
            [[fallthrough]];

        // RGB
        case PixelFormat::RGB:
            _pixelBuffer[index] = color.red;
            _pixelBuffer[index + 1] = color.green;
            _pixelBuffer[index + 2] = color.blue;
            break;

        default:
            break;
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      ピクセルを設定（グレースケール）
 *  \param[in]	x			X座標
 *  \param[in]	y			Y座標
 *  \param[in]	grayColor	輝度
 *  \param[in]	alpha		アルファ（省略時は0xFF，無透過フォーマットの場合は無視）
 */
/* ------------------------------------------------------------------------- */
void Bitmap::SetPixelGray(uint32_t x, uint32_t y, uint8_t grayColor, uint8_t alpha) noexcept
{
    if (!_isValid || !IsInCanvas(x, y))
    {
        return;
    }

    auto index = static_cast<size_t>((y * _width) + x) * _pixelSize;

    switch (_pixelFormat)
    {
        // グレースケール＋アルファ
        case PixelFormat::GrayAlpha:
            _pixelBuffer[index + 1] = alpha;
            [[fallthrough]];

        // グレースケール
        case PixelFormat::Gray:
            _pixelBuffer[index] = grayColor;
            break;

        // RGBA
        case PixelFormat::RGBA:
            _pixelBuffer[index + 3] = alpha;
            [[fallthrough]];

        // RGB
        case PixelFormat::RGB:
            _pixelBuffer[index] = grayColor;
            _pixelBuffer[index + 1] = grayColor;
            _pixelBuffer[index + 2] = grayColor;
            break;

        default:
            break;
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      ピクセルの取得
 *  \param[in]  x   X座標
 *  \param[in]  y   Y座標
 *  \return     座標に対応した色
 */
/* ------------------------------------------------------------------------- */
Color Bitmap::GetPixel(uint32_t x, uint32_t y) const noexcept
{
    if (!IsInCanvas(x, y))
    {
        return Color();
    }

    auto index = static_cast<size_t>((y * _width) + x) * _pixelSize;

    Color color;

    switch (_pixelFormat)
    {
        // グレースケール＋アルファ
        case PixelFormat::GrayAlpha:
            color.red   = _pixelBuffer[index];
            color.green = color.red;
            color.blue  = color.red;
            color.alpha = _pixelBuffer[index + 1];
            break;

        // グレースケール
        case PixelFormat::Gray:
            color.red   = _pixelBuffer[index];
            color.green = color.red;
            color.blue  = color.red;
            color.alpha = 255;
            break;

        // RGBA
        case PixelFormat::RGBA:
            color.red   = _pixelBuffer[index];
            color.green = _pixelBuffer[index + 1];
            color.blue  = _pixelBuffer[index + 2];
            color.alpha = _pixelBuffer[index + 3];
            break;

        // RGB
        case PixelFormat::RGB:
            color.red   = _pixelBuffer[index];
            color.green = _pixelBuffer[index + 1];
            color.blue  = _pixelBuffer[index + 2];
            color.alpha = 255;
            break;

        default:
            break;

    }

    return color;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      塗りつぶし
 *  \param[in]  x       塗りつぶし範囲の左上のX座標
 *  \param[in]  y       塗りつぶし範囲の左上のY座標
 *  \param[in]  width   塗りつぶし範囲の幅
 *  \param[in]  height  塗りつぶし範囲の高さ
 *  \param[in]  color   塗りつぶす色
 */
/* ------------------------------------------------------------------------- */
void Bitmap::Fill(uint32_t x, uint32_t y, uint32_t width, uint32_t height, const Color &color) noexcept
{
    if ((_pixelFormat == PixelFormat::Gray) || (_pixelFormat == PixelFormat::GrayAlpha))
    {
        FillGray(x, y, width, height, color.ToGrayScale(), color.alpha);
    }

    auto right = x + width;
    auto bottom = y + height;

    for (auto px = x; px < right; px++)
    {
        for (auto py = y; py < bottom; py++)
        {
            SetPixel(px, py, color);
        }
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      塗りつぶし（グレースケール）
 *  \param[in]  x           塗りつぶし範囲の左上のX座標
 *  \param[in]  y           塗りつぶし範囲の左上のY座標
 *  \param[in]  width       塗りつぶし範囲の幅
 *  \param[in]  height      塗りつぶし範囲の高さ
 *  \param[in]  color       塗りつぶす色
 *  \param[in]  grayColor   グレースケールの色（輝度）
 *  \param[in]  alpha       アルファ値
 */
/* ------------------------------------------------------------------------- */
void Bitmap::FillGray(uint32_t x, uint32_t y, uint32_t width, uint32_t height, uint8_t grayColor, uint8_t alpha) noexcept
{
    auto right = x + width;
    auto bottom = y + height;

    for (auto px = x; px < right; px++)
    {
        for (auto py = y; py < bottom; py++)
        {
            SetPixelGray(px, py, grayColor, alpha);
        }
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      他のビットマップの内容をコピー
 *  \param[in]  x         X座標
 *  \param[in]  y         Y座標
 *  \param[in]  source    コピー元のビットマップ
 *  \param[in]  sourceX   コピー元のX座標
 *  \param[in]  sourceY   コピー元のY座標
 *  \param[in]  width     コピーする幅
 *  \param[in]  height    コピーする高さ
 *
 */
/* ------------------------------------------------------------------------- */
void Bitmap::Copy(uint32_t x, uint32_t y, const Bitmap &source, uint32_t sourceX, uint32_t sourceY, uint32_t width, uint32_t height) noexcept
{
    if (!IsInCanvas(x, y) || !source.IsInCanvas(sourceX, sourceY))
    {
        return;
    }

    // コピー元の矩形が範囲外にならないよう補正
    if ((sourceX + width) > source.GetWidth())
    {
        width = source.GetWidth() - sourceX;
    }
    if ((sourceY + height) > source.GetHeight())
    {
        height = source.GetHeight() - sourceY;
    }

    // コピー先の矩形が範囲外にならないよう補正
    if ((x + width) > _width)
    {
        width = _width - x;
    }
    if ((y + height) > _height)
    {
        height = _height - y;
    }

    // ピクセルフォーマットが同じなら行単位でコピー
    if (_pixelFormat == source._pixelFormat)
    {
        for (uint32_t offsetY = 0; offsetY < height; offsetY++)
        {
            size_t size = _pixelSize * static_cast<size_t>(width);
            auto *destAddress = GetPixelAddress(x, y + offsetY);
            if (destAddress == nullptr)
            {
                break;
            }

            auto *srcAddress = source.GetPixelAddress(sourceX, sourceY + offsetY);
            if (srcAddress == nullptr)
            {
                break;
            }

            memcpy(destAddress, srcAddress, size);
        }
    }
    // ピクセルフォーマットが異なる場合はピクセル単位で変換しながらコピー
    else
    {
        for (uint32_t offsetY = 0; offsetY < height; offsetY++)
        {
            for (uint32_t offsetX = 0; offsetX < width; offsetX++)
            {
                SetPixel(x + offsetX, y + offsetY, source.GetPixel(sourceX + offsetX, sourceY + offsetY));
            }
        }
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      指定したピクセルのアドレスを取得
 *  \param[in]  x   X座標
 *  \param[in]  y   Y座標
 *  \return     座標に対応したアドレス
 */
/* ------------------------------------------------------------------------- */
uint8_t *Bitmap::GetPixelAddress(uint32_t x, uint32_t y) const noexcept
{
    if (!IsInCanvas(x, y))
    {
        return nullptr;
    }

    auto index = static_cast<size_t>(_width * y + x) * _pixelSize;

    return &_pixelBuffer[index];
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      PNG形式から読み込み
 *  \param[in]	path	読み込むファイルのパス
 *  \retval		true	成功
 *  \retval		false	失敗
 */
/* ------------------------------------------------------------------------- */
bool Bitmap::LoadPNG(const std::filesystem::path &path) noexcept
{
    if (_isValid)
    {
        return false;
    }

    png_image png = {};
    png.version = PNG_IMAGE_VERSION;

    // PNGファイルの読み込み
    MTK::File::Pointer fp(path, "rb");
    if (!fp)
    {
        return false;
    }
    png_image_begin_read_from_stdio(&png, fp);
    if (PNG_IMAGE_FAILED(png))
    {
        return false;
    }

    // 読み込んだPNGがパレット形式であればSimplified APIでは読み込めないため，
    // 別の関数に処理を移す．
    if (png.format & PNG_FORMAT_FLAG_COLORMAP)
    {
        png_image_free(&png);
        return LoadColormappedPNG(path);
    }

    // ピクセルフォーマットを取得
    PixelFormat pixelFormat = PixelFormat::RGBA;
    switch (png.format)
    {
        case PNG_FORMAT_GRAY:
            pixelFormat = PixelFormat::Gray;
            break;

        case PNG_FORMAT_GA:
            pixelFormat = PixelFormat::GrayAlpha;
            break;

        case PNG_FORMAT_RGB:
            pixelFormat = PixelFormat::RGB;
            break;

        case PNG_FORMAT_RGBA:
            pixelFormat = PixelFormat::RGBA;
            break;

        default:
            png_image_free(&png);
            return false;
    };

    // ビットマップを初期化
    Initialize(static_cast<uint32_t>(png.width), static_cast<uint32_t>(png.height), pixelFormat);

    // バッファサイズをチェック
    auto stride = PNG_IMAGE_ROW_STRIDE(png);
    if (_pixelBufferSize != PNG_IMAGE_BUFFER_SIZE(png, stride))
    {
        png_image_free(&png);
        Reset();
        return false;
    }

    // ピクセル情報を読み込み
    png_image_finish_read(&png, nullptr, _pixelBuffer.get(), static_cast<png_int_32>(stride), nullptr);
    if (PNG_IMAGE_FAILED(png))
    {
        png_image_free(&png);
        Reset();
        return false;
    }

    png_image_free(&png);

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      パレット方式のPNG画像の読み込み
 *  \param[in]	path	読み込むファイルのパス
 *  \retval     true    成功
 *  \retval     false   失敗
 */
/* ------------------------------------------------------------------------- */
bool Bitmap::LoadColormappedPNG(const std::filesystem::path &path) noexcept
{
    constexpr size_t kSignatureSize = 8;    // PNGのシグネチャのサイズ（8バイト固定）

    // ファイルをオープン
    MTK::File::Pointer fp(path, "rb");
    if (!fp)
    {
        return false;
    }

    png_structp png = nullptr;
    png_infop info = nullptr;

    // PNGの初期化
    try
    {
        // ファイルがPNGかどうかをチェック
        unsigned char signature[kSignatureSize];
        fread(signature, 1, kSignatureSize, fp);
        if (png_sig_cmp(signature, 0, kSignatureSize))
        {
            throw MTK::Format("{} is not png.", {path});
        }

        // 読み込み用構造体を生成
        png = png_create_read_struct(PNG_LIBPNG_VER_STRING, nullptr, nullptr, nullptr);
        if (png == nullptr)
        {
            throw std::string("failed to initialize png_structp.");
        }

        // PNG情報の構造体を生成
        info = png_create_info_struct(png);
        if (info == nullptr)
        {
            throw std::string("failed to initialize png_infop.");
        }

        // PNGファイルの読み込み
        png_init_io(png, fp);
        png_set_sig_bytes(png, kSignatureSize);

        // PNGファイルの内容を読み込む
        // デコードはここで行われるらしい
        png_read_png(png, info, PNG_TRANSFORM_STRIP_16 | PNG_TRANSFORM_PACKING, nullptr);

        // パレット形式のPNGファイルでなければこの関数では扱わない
        if (png_get_color_type(png, info) != PNG_COLOR_TYPE_PALETTE)
        {
            throw MTK::Format("{} is not colormapped png.", {path});
        }
    }
    // 上記までのエラー処理
    catch (const std::string errorMessage)
    {
        PutError(MsgID::PNGReadError, {errorMessage});

        if ((png != nullptr) && (info != nullptr))
        {
            png_destroy_read_struct(&png, &info, nullptr);
        }

        return false;
    }

    // PNGファイルの情報を元にビットマップを初期化
    Initialize(png_get_image_width(png, info), png_get_image_height(png, info), PixelFormat::RGBA);

    // パレット情報を取得
    png_colorp pallet = nullptr;
    int palletCount = 0;
    png_get_PLTE(png, info, &pallet, &palletCount);

    // トランス情報を取得
    // パレット情報は色のみでアルファチャンネルが存在しないため，別途取得する必要がある
    png_bytep trans = nullptr;
    int transCount = 0;
    png_get_tRNS(png, info, &trans, &transCount, nullptr);

    // 画像データを取得
    // データは行単位で，パレット形式の場合はパレットのインデックスが格納されている
    auto rows = png_get_rows(png, info);

    // 各ピクセルの色を取得して適用
    for (uint32_t y = 0; y < _height; y++)
    {
        for (uint32_t x = 0; x < _width; x++)
        {
            Color color;
            size_t index = rows[y][x];

            // パレットを参照して色を取得
            if (index < static_cast<size_t>(palletCount))
            {
                auto pngColor = pallet[index];
                color.Set(pngColor.red, pngColor.green, pngColor.blue);
            }

            // トランスを取得してアルファ値を取得
            if (index < static_cast<size_t>(transCount))
            {
                color.alpha = trans[index];
            }

            // ピクセルを設定
            SetPixel(x, y, color);
        }
    }

    // 後処理
    png_destroy_read_struct(&png, &info, nullptr);

    return false;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      PNG形式で保存
 *  \param[in]	path	出力ファイルのパス
 *  \retval		true	成功
 *  \retval		false	失敗
 */
/* ------------------------------------------------------------------------- */
bool Bitmap::SavePNG(const std::filesystem::path &path) noexcept
{
    png_image png = {};
    png.version = PNG_IMAGE_VERSION;

    png.width = static_cast<png_uint_32>(_width);
    png.height = static_cast<png_uint_32>(_height);

    switch (_pixelFormat)
    {
        // グレースケール
        case PixelFormat::Gray:
            png.format = PNG_FORMAT_GRAY;
            break; 

        // グレースケール＋アルファ
        case PixelFormat::GrayAlpha:
            png.format = PNG_FORMAT_GA;
            break;

        // RGB
        case PixelFormat::RGB:
            png.format = PNG_FORMAT_RGB;
            break;

        // RGBA
        case PixelFormat::RGBA:
            png.format = PNG_FORMAT_RGBA;
            break;

        default:
            return false;
    }

    MTK::File::Pointer fp(path, "wb");
    if (!fp)
    {
        return false;
    }
    png_image_write_to_stdio(&png, fp, 0, _pixelBuffer.get(), static_cast<png_int_32>(PNG_IMAGE_ROW_STRIDE(png)), nullptr);

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      このクラスを初期状態に戻す
 */
/* ------------------------------------------------------------------------- */
void Bitmap::Reset() noexcept
{
    _isValid = false;
    _pixelBuffer.reset();
    _pixelBufferSize = 0;
    _pixelSize = 0;
    _width = _height = 0;
}

}	// namespace texatlas

// vim: et ts=4 sw=4 sts=4


