/* ------------------------------------------------------------------------- */
/*!
 *  \file       bitmap.h
 *  \brief      ビットマップクラス
 *  \date       Since: September 18, 2018. 22:45:25 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_BITMAP_H_1537278325
#define INCGUARD_BITMAP_H_1537278325

#include <filesystem>
#include <memory>
#include <stddef.h>
#include <stdint.h>

#include "color.h"

namespace texatlas
{
//! ビットマップクラス
class Bitmap
{
public:
    //! ピクセルフォーマット
    enum class PixelFormat
    {
        Gray,			//!< グレースケール
        GrayAlpha,		//!< グレースケール＋アルファ
        RGB,			//!< RGB
        RGBA			//!< RGBA
    };


    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ
     */
    /* ------------------------------------------------------------------------- */
    constexpr Bitmap() noexcept
        : _isValid(false)
        , _pixelFormat(PixelFormat::Gray)
        , _pixelSize(0)
        , _pixelBuffer(nullptr)
        , _pixelBufferSize(0)
        , _width(0)
        , _height(0)
    {}

    Bitmap(uint32_t width, uint32_t height, PixelFormat pixelFormat) noexcept;

    bool Initialize(uint32_t width, uint32_t height, PixelFormat pixelFormat) noexcept;

    void SetPixel(uint32_t x, uint32_t y, const Color &color) noexcept;
    void SetPixelGray(uint32_t x, uint32_t y, uint8_t grayColor, uint8_t alpha = 0xFF) noexcept;

    Color GetPixel(uint32_t x, uint32_t y) const noexcept;

    void Fill(uint32_t x, uint32_t y, uint32_t width, uint32_t height, const Color &color) noexcept;
    void FillGray(uint32_t x, uint32_t y, uint32_t width, uint32_t height, uint8_t grayColor, uint8_t alpha) noexcept;

    void Copy(uint32_t x, uint32_t y, const Bitmap &source, uint32_t sourceX, uint32_t sourceY, uint32_t sourceWidth, uint32_t sourceHeight) noexcept;


    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      ピクセルバッファを取得
     *  \return		ピクセルバッファ
     */
    /* ------------------------------------------------------------------------- */
    void *GetPixelBuffer() const noexcept
    {
        return _pixelBuffer.get();
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      ピクセルバッファのサイズを取得
     *  \return		ピクセルバッファのサイズ
     */
    /* ------------------------------------------------------------------------- */
    constexpr size_t GetPixelBufferSize() const noexcept
    {
        return _pixelBufferSize;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      幅を取得
     *  \return     幅
     */
    /* ------------------------------------------------------------------------- */
    constexpr uint32_t GetWidth() const noexcept
    {
        return static_cast<uint32_t>(_width);
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      高さを取得
     *  \return     高さ
     */
    /* ------------------------------------------------------------------------- */
    constexpr uint32_t GetHeight() const noexcept
    {
        return static_cast<uint32_t>(_height);
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      指定座標がキャンバス内に入っているかをチェック
     *  \retval     true    入っている
     *  \retval     false   入っていない
     */
    /* ------------------------------------------------------------------------- */
    constexpr bool IsInCanvas(uint32_t x, uint32_t y) const noexcept
    {
        return (x < _width) && (y < _height);
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      boolへのキャスト
     *  \retval     true    このクラスが有効
     *  \retval     false   このクラスは無効
     */
    /* ------------------------------------------------------------------------- */
    explicit constexpr operator bool() const noexcept
    {
        return _isValid;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      否定演算子
     *  \retval     true    このクラスは無効
     *  \retval     false   このクラスが有効
     */
    /* ------------------------------------------------------------------------- */
    constexpr bool operator!() const noexcept
    {
        return !static_cast<bool>(*this);
    }

    bool LoadPNG(const std::filesystem::path &path) noexcept;
    bool SavePNG(const std::filesystem::path &path) noexcept;

    void Reset() noexcept;

private:
    bool LoadColormappedPNG(const std::filesystem::path &path) noexcept;

    uint8_t *GetPixelAddress(uint32_t x, uint32_t y) const noexcept;

    bool _isValid;			                    //!< 有効フラグ
    PixelFormat _pixelFormat;		            //!< ピクセルフォーマット
    size_t _pixelSize;					        //!< 1ピクセルあたりのバイト数
    std::unique_ptr<uint8_t[]> _pixelBuffer;    //!< ピクセルバッファ
    size_t _pixelBufferSize;		            //!< ピクセルバッファのサイズ
    uint32_t _width;						    //!< 幅
    uint32_t _height;					        //!< 高さ
};
}	// namespace texatlas

#endif  // INCGUARD_BITMAP_H_1537278325

// vim: et ts=4 sw=4 sts=4


