/* ------------------------------------------------------------------------- */
/*!
 *  \file       color.h
 *  \brief      色定義
 *  \date       Since: September 24, 2018. 18:10:45 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_COLOR_H_1537780245
#define INCGUARD_COLOR_H_1537780245

#include <stdint.h>
#include <math.h>

namespace texatlas
{
//! 色定義
struct Color
{
    uint8_t red;   //!< 赤
    uint8_t green; //!< 緑
    uint8_t blue;  //!< 青
    uint8_t alpha; //!< アルファ値
 

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ
     */
    /* ------------------------------------------------------------------------- */
    constexpr Color() noexcept
        : red(0)
        , green(0)
        , blue(0)
        , alpha(0)
    {}


    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      コンストラクタ
     *  \param[in]  inRed   赤成分
     *  \param[in]  inGreen 緑成分
     *  \param[in]  inBlue  青成分
     *  \param[in]  inAlpha アルファ成分
     */
    /* ------------------------------------------------------------------------- */
    Color(uint8_t inRed, uint8_t inGreen, uint8_t inBlue, uint8_t inAlpha = 0xFF) noexcept
    {
        Set(inRed, inGreen, inBlue, inAlpha);
    }


    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      色の設定
     *  \param[in]  inRed   赤成分
     *  \param[in]  inGreen 緑成分
     *  \param[in]  inBlue  青成分
     *  \param[in]  inAlpha アルファ成分
     */
    /* ------------------------------------------------------------------------- */
    void Set(uint8_t inRed, uint8_t inGreen, uint8_t inBlue, uint8_t inAlpha = 0xFF) noexcept
    {
        red = inRed;
        green = inGreen;
        blue = inBlue;
        alpha = inAlpha;
    }


    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      グレースケール変換
     *  \param[in]  gamma   ガンマ値（デフォルトで2.2）
     *  \return     輝度
     *  \note
     *      ガンマ補正＋輝度変換．速度は気にしない．
     */
    /* ------------------------------------------------------------------------- */
    uint8_t ToGrayScale(float gamma = 2.2f) const noexcept
    {
        float r = static_cast<float>(red) / 255.0f;
        float g = static_cast<float>(green) / 255.0f;
        float b = static_cast<float>(blue) / 255.0f;

        r = powf(r, gamma) * 0.2126f;
        g = powf(g, gamma) * 0.7152f;
        b = powf(b, gamma) * 0.0722f;

        float y = powf((r + g + b), (1.0f / gamma));

        return static_cast<uint8_t>(y * 255.0f);
    }
};
}   // namespace texatlas

#endif  // INCGUARD_COLOR_H_1537780245

// vim: et ts=4 sw=4 sts=4
