/* ------------------------------------------------------------------------- */
/*!
 *  \file       image_file_info.cc
 *  \brief      画像ファイル情報
 *  \date       Since: June 6, 2021. 17:44:46 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include "image_file_info.h"

#include <mtk/csv_parser.h>
#include <mtk/file_pointer.h>
#include <mtk/text_hash.h>
#include <mtk/text_converter.h>
#include <mtk/utility.h>
#include <png.h>

#include "message.h"

namespace texatlas
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      コンストラクタ
 *  \param[in]  path        画像ファイルのパス
 *  \param[in]  hashSeed    ハッシュのシード値
 */
/* ------------------------------------------------------------------------- */
ImageFileInfo::ImageFileInfo(const std::filesystem::path &path, uint32_t hashSeed) noexcept
    : _isValid(false)
    , _path()
    , _name()
    , _hash(0)
    , _width(0)
    , _height(0)
{
    Load(path, hashSeed);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      読み込み
 *  \param[in]  path        読み込むパス
 *  \param[in]  hashSeed    ハッシュシード値
 *  \retval     true        成功
 *  \retval     false       失敗
 */
/* ------------------------------------------------------------------------- */
bool ImageFileInfo::Load(const std::filesystem::path &path, uint32_t hashSeed) noexcept
{
    bool result = false;
    _path = path;

    // ファイルの拡張子を小文字で取得
    auto filetype = path.extension().u8string();
    for (auto &c : filetype)
    {
        if ((c >= 'A') && (c <= 'Z'))
        {
            c -= 'a' - 'A';
        }
    }

    // 拡張子がpngならPNG画像の読み込み
    if (filetype == u8".png")
    {
        result = LoadPNG(path);
    }
    // ...以降，他の画像形式に対応するならここに追加する

    // 読み込みに成功していたら情報を構築
    if (result)
    {
        _name = std::filesystem::path(path).replace_extension("").u8string();
#if defined(MTK_TARGET_WINDOWS)
        MTK::Utility::ReplaceString(_name, "\\", "/");
#endif
        _hash = MTK::Hash::FNV1a(_name.c_str(), hashSeed);
        _isValid = true;
        LoadChildren(hashSeed);
    }

    return _isValid;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      PNG画像から情報を読み込み
 *  \param[in]  path    読み込むパス
 *  \retval     true    成功
 *  \retval     false   失敗
 */
/* ------------------------------------------------------------------------- */
bool ImageFileInfo::LoadPNG(const std::filesystem::path &path) noexcept
{
    png_image png = {};
    png.version = PNG_IMAGE_VERSION;

    MTK::File::Pointer fp(path, "rb");
    if (!fp)
    {
        return false;
    }

    png_image_begin_read_from_stdio(&png, fp);
    if (PNG_IMAGE_FAILED(png))
    {
        return false;
    }

    _width = static_cast<uint32_t>(png.width);
    _height = static_cast<uint32_t>(png.height);

    png_image_free(&png);

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      子の位置情報を読み込み
 *  \param[in]  hashSeed    ハッシュシード値
 *  \return     子の位置情報の数
 */
/* ------------------------------------------------------------------------- */
size_t ImageFileInfo::LoadChildren(uint32_t hashSeed) noexcept
{
    MTK::CSVParser csv;

    auto csvPath = _path;
    csvPath.replace_extension(".csv");
    if (csv.Load(csvPath))
    {
        for (int record = 0; record < csv.GetRecordCount(); record++)
        {
            try
            {
                Child child;
                auto name = csv.GetField(record, 0);
                if (name.empty())
                {
                    continue;
                }
                child.name = _name + "/" + name;
                child.hash = MTK::Hash::FNV1a(child.name.c_str(), hashSeed);
                child.x = static_cast<uint32_t>(std::stoul(csv.GetField(record, 1)));
                child.y = static_cast<uint32_t>(std::stoul(csv.GetField(record, 2)));
                child.width = static_cast<uint32_t>(std::stoul(csv.GetField(record, 3)));
                child.height = static_cast<uint32_t>(std::stoul(csv.GetField(record, 4)));
                _children.push_back(child);
            }
            catch ([[maybe_unused]] std::exception e)
            {
                PutError(MsgID::InvalidCSVRecord, {csv.GetField(record, 0)});
                continue;
            }
        }
    }

    return _children.size();
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      情報を表示
 */
/* ------------------------------------------------------------------------- */
void ImageFileInfo::Print() const noexcept
{
#if defined(MTK_TARGET_WINDOWS)
    std::string path;
    if (!MTK::ToShiftJIS(path, _path.u8string().c_str(), _path.u8string().size(), MTK::TextEncoding::UTF8))
    {
        return;
    }
#else
    auto path = _path;
#endif

    if (!_isValid)
    {
        printf("%s is not available.\n", path.c_str());
    }
    else
    {
        printf("|--%s\n", path.c_str());
        printf("|  |--name: %s\n", _name.c_str());
        printf("|  |--hash: 0x%08X\n", _hash);
        printf("|  |--size: %dx%d\n", _width, _height);

        for (const auto &child : _children)
        {
            printf("|  |--%s\n", child.name.c_str());
            printf("|  |  |--hash: 0x%08X\n", child.hash);
            printf("|  |  |--pos:  %d,%d\n", child.x, child.y);
            printf("|  |  |--size: %dx%d\n", child.width, child.height);
        }
    }
}

}   // namespace texatlas

// vim: et ts=4 sw=4 sts=4
