/* ------------------------------------------------------------------------- */
/*!
 *  \file       image_file_info.h
 *  \brief      画像ファイル情報
 *  \date       Since: June 6, 2021. 17:44:48 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_IMAGE_FILE_INFO_H_1622969088
#define INCGUARD_IMAGE_FILE_INFO_H_1622969088

#include <cstdint>
#include <filesystem>
#include <string>
#include <vector>

namespace texatlas
{
class ImageFileInfo
{
public:
    struct Child
    {
        std::string name;
        uint32_t hash;
        uint32_t x;
        uint32_t y;
        uint32_t width;
        uint32_t height;
    };

    ImageFileInfo(const std::filesystem::path &path, uint32_t hashSeed) noexcept;

    bool Load(const std::filesystem::path &path, uint32_t hashSeed) noexcept;

    void Print() const noexcept;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      boolへのキャスト
     *  \retval     true    このクラスが有効
     *  \retval     false   このクラスは無効
     */
    /* ------------------------------------------------------------------------- */
    explicit constexpr operator bool() const noexcept
    {
        return _isValid;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      否定演算子
     *  \retval     true    このクラスは無効
     *  \retval     false   このクラスが有効
     */
    /* ------------------------------------------------------------------------- */
    constexpr bool operator!() const noexcept
    {
        return !static_cast<bool>(*this);
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      パスを取得
     *  \return     対象ファイルのパス
     */
    /* ------------------------------------------------------------------------- */
    constexpr const std::filesystem::path &GetPath() const noexcept
    {
        return _path;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      画像の幅を取得
     *  \return     画像の幅
     */
    /* ------------------------------------------------------------------------- */
    constexpr uint32_t GetWidth() const noexcept
    {
        return _width;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      画像の高さを取得
     *  \return     画像の高さ
     */
    /* ------------------------------------------------------------------------- */
    constexpr uint32_t GetHeight() const noexcept
    {
        return _height;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      ハッシュ値を取得
     *  \return     ハッシュ値
     */
    /* ------------------------------------------------------------------------- */
    constexpr uint32_t GetHash() const noexcept
    {
        return _hash;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      名前を取得
     *  \return     名前
     */
    /* ------------------------------------------------------------------------- */
    constexpr const std::string &GetName() const noexcept
    {
        return _name;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      子の位置情報を取得
     *  \return     子の位置情報の配列
     */
    /* ------------------------------------------------------------------------- */
    constexpr const std::vector<Child> &GetChildren() const noexcept
    {
        return _children;
    }

private:
    bool LoadPNG(const std::filesystem::path &path) noexcept;
    
    size_t LoadChildren(uint32_t hashSeed) noexcept;

    bool _isValid;
    std::filesystem::path _path;
    std::string _name;
    uint32_t _hash;
    uint32_t _width;
    uint32_t _height;

    std::vector<Child> _children;
};

}   // namespace texatlas

#endif	// INCGUARD_IMAGE_FILE_INFO_H_1622969088

// vim: et ts=4 sw=4 sts=4
