// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       config_parser.cc
 *  \brief      設定データ読み込みクラス
 *  \date       Since: October 6, 2018. 23:09:01 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <mtk/config_parser.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

namespace MTK
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      設定データの読み込み
 *  \param[in]  path    読み込むファイルのパス
 *  \retval     true    成功
 *  \retval     false   失敗
 */
/* ------------------------------------------------------------------------- */
bool ConfigParser::Load(const std::filesystem::path &path) noexcept
{
    // TextStreamに読み込み
    TextStream textStream;
    if (!textStream.Load(path))
    {
        return false;
    }

    // テキストを解析
    bool result = ParseText(textStream);

    return result;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      コマンドライン引数からパース
 *  \param[in]  argc        コマンドライン引数の数
 *  \param[in]  argv        コマンドライン引数
 *  \param[in]  parseStart  パースを開始する引数のインデックス
 *  \retval     true        成功
 *  \retval     false       失敗
 */
/* ------------------------------------------------------------------------- */
bool ConfigParser::ParseArgs(int argc, const char *argv[], int parseStart) noexcept
{
    for (int i = parseStart; i < argc; i++)
    {
        // その引数がキーとして認識できなければ飛ばす
        if (!IsArgsOptionKey(argv[i]))
        {
            continue;
        }

        // 論理値キーかどうかを取得
        bool isBooleanKey = false;
        if (i == (argc - 1))
        {
            isBooleanKey = true;
        }
        else if (IsArgsOptionKey(argv[i + 1]))
        {
            isBooleanKey = true;
        }

        // 引数のキーをセクション名とキー名に分割
        std::string argString(&argv[i][2]);
        std::string section, key;
        auto colonPos = argString.find(":");
        if (colonPos == std::string::npos)
        {
            section.clear();
            key = &argv[i][2];
        }
        else
        {
            section = argString.substr(0, colonPos);
            key = &argString.data()[colonPos + 1];
        }

        // 解析させるために行として生成
        std::string line;
        line += key;
        line += "=";
        line += (isBooleanKey) ? "true" : argv[i + 1];

        // 行をCharacterArrayに変換してパーサーに解析させる．
        // ちょっと非効率だけど，パース処理を共通化させたいので．
        CharacterArray charaArray;
        TextStream stream(reinterpret_cast<const uint8_t*>(line.data()), line.size());
        stream.Read(charaArray);
        TokenArray tokenArray;
        if (!ParseLine(tokenArray, charaArray))
        {
            fprintf(stderr, "parser error: command line: %s\n", argv[i]);
            return false;
        }
        if (!ParseToken(tokenArray, section))
        {
            fprintf(stderr, "parser error: command line: %s\n", argv[i]);
            return false;
        }

        // 論理値キーでなければ2つ進める
        if (!isBooleanKey)
        {
            i++;
        }
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      設定値をクリア
 */
/* ------------------------------------------------------------------------- */
void ConfigParser::Clear() noexcept
{
    _currentSection.clear();
    _valueArray.clear();
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      キーが存在するかを取得
 *  \param[in]  key     チェックするキー
 *  \retval     true    存在する
 *  \retval     false   存在しない
 */
/* ------------------------------------------------------------------------- */
bool ConfigParser::Exists(const char *key) const noexcept
{
    const auto it = _valueArray.find(key);
    return (it != _valueArray.end());
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      設定値を文字列で取得
 *  \param[in]  key             設定値のキー
 *  \param[in]  defaultValue    デフォルト値
 *  \return     キーに対応した文字列
 */
/* ------------------------------------------------------------------------- */
std::string ConfigParser::GetString(const char *key, const char *defaultValue) const noexcept
{
    const auto it = _valueArray.find(key);
    if (it != _valueArray.end())
    {
        return it->second.name;
    }

    std::string str;
    if (defaultValue != nullptr)
    {
        str = defaultValue;
    }

    return str;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      設定値を整数で取得
 *  \param[in]  key             設定値のキー
 *  \param[in]  defaultValue    デフォルト値
 *  \return     キーに対応した整数値
 */
/* ------------------------------------------------------------------------- */
int32_t ConfigParser::GetInt(const char *key, int32_t defaultValue) const noexcept
{
    const auto it = _valueArray.find(key);
    if (it != _valueArray.end())
    {
        const auto &token = it->second;

        if (token.type == TokenType::IntValue)
        {
            return token.intValue;
        }
        else if (token.type == TokenType::FloatValue)
        {
            return static_cast<int32_t>(token.floatValue);
        }
    }

    return defaultValue;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      設定値を少数点数で取得
 *  \param[in]  key             設定値のキー
 *  \param[in]  defaultValue    デフォルト値
 *  \return     キーに対応した小数点数値
 */
/* ------------------------------------------------------------------------- */
float ConfigParser::GetFloat(const char *key, float defaultValue) const noexcept
{
    const auto it = _valueArray.find(key);
    if (it != _valueArray.end())
    {
        const auto &token = it->second;

        if (token.type == TokenType::IntValue)
        {
            return static_cast<float>(token.intValue);
        }
        else if (token.type == TokenType::FloatValue)
        {
            return token.floatValue;
        }
    }

    return defaultValue;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      設定値を論理値で取得
 *  \param[in]  key             設定値のキー
 *  \param[in]  defaultValue    デフォルト値
 *  \return     キーに対応した論理値
 */
/* ------------------------------------------------------------------------- */
bool ConfigParser::GetBool(const char *key, bool defaultValue) const noexcept
{
    const auto it = _valueArray.find(key);
    if (it != _valueArray.end())
    {
        const auto &token = it->second;

        if (token.type == TokenType::IntValue)
        {
            return (token.intValue != 0);
        }
        else if (token.type == TokenType::FloatValue)
        {
            return (token.floatValue != 0.0f);
        }
    }

    return defaultValue;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          テキストからパース
 *  \param[in,out]  textStream  パースするテキストのストリーム
 *  \retval         true        成功
 *  \retval         false       失敗
 */
/* ------------------------------------------------------------------------- */
bool ConfigParser::ParseText(TextStream &textStream) noexcept
{
    CharacterArray line;

    for (int lineCount = 1;; lineCount++)
    {
        // 終端に達していれば終了
        if (textStream.IsTerminated())
        {
            break;
        }

        // 1行読み込み
        line.clear();
        textStream.ReadLine(line);
        if (line.size() == 0)
        {
            continue;
        }

        // 行をトークンに変換
        TokenArray tokenArray;
        if (!ParseLine(tokenArray, line))
        {
            fprintf(stderr, "parser error: line %d\n", lineCount);
            return false;
        }

        // トークンを解析
        if (!ParseToken(tokenArray, _currentSection))
        {
            fprintf(stderr, "parser error: line %d\n", lineCount);
            return false;
        }

#if 0
        // トークンの中身を確認
        for (const auto &token : tokenArray)
        {
            static const char *s_TokenType[] =
            {
                "String",
                "IntValue",
                "FloatValue",
                "SectionName",
                "Equal",
            };
            printf("%d: %s: \"%s\"\n", lineCount, s_TokenType[static_cast<int>(token.type)], token.name.c_str());
        }
#endif
    }


#if 0
    // 内容確認
    printf("size: %lu\n", _valueArray.size());
    for (const auto it : _valueArray)
    {
        printf("%s: %s\n", it.first.c_str(), it.second.name.c_str());
    }
#endif
    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      行をトークンにパース
 *  \param[out] tokenArray  トークン配列の格納先
 *  \param[in]  line        行の文字配列
 *  \retval     true        成功
 *  \retval     false       失敗
 */
/* ------------------------------------------------------------------------- */
bool ConfigParser::ParseLine(ConfigParser::TokenArray &tokenArray, const CharacterArray &line) noexcept
{
    auto arraySize = line.size();

    // 行を1文字づつ走査
    for (
        size_t index = 0, nextIndex = index + 1;
        index < arraySize;
        index = nextIndex
    )
    {
        nextIndex = index + 1;
        Token token;
        bool isForcedString = false;

        // スペースは飛ばす
        if (line[index].IsSpace())
        {
            index = nextIndex;
            continue;
        }
        // '#'があればそれ以降は飛ばす
        else if (line[index].GetCode() == '#')
        {
            return true;
        }
        // '['で始まっていればセクション名
        else if (line[index].GetCode() == '[')
        {
            // ']'を探す．見つからなければエラー
            int endPos = line.Find(']', static_cast<uint32_t>(index + 1));
            if (endPos <= 0)
            {
                 return false;
            }

            token.name = line.ToString(static_cast<uint32_t>(index + 1), static_cast<uint32_t>(endPos));
            token.type = TokenType::SectionName;
            nextIndex = static_cast<size_t>(endPos) + 1;
        }
        // '"'で始まっていれば文字列
        else if (line[index].GetCode() == '"')
        {
            // 終わりの'"'を探す．見つからなければエラー
            int endPos = line.Find('"', static_cast<uint32_t>(index + 1));
            if (endPos <= 0)
            {
                return false;
            }

            token.name = line.ToString(static_cast<uint32_t>(index + 1), static_cast<uint32_t>(endPos));
            token.type = TokenType::String;
            nextIndex = static_cast<size_t>(endPos) + 1;
            isForcedString = true;
        }
        // 数値で始まるトークンの解析
        else if (line[index].IsDigit() || (line[index].GetCode() == '-'))
        {
            // まずは整数と仮定する
            token.type = TokenType::IntValue;

            token.name += line[index].ToString();

            // 1文字づつ走査
            for (size_t i = 1;; i++)
            {
                const auto &character = line[index + i];

                // 終了コードか'='が来ていれば終了
                if (character.IsNull() || (character.GetCode() == '='))
                {
                    nextIndex = index + i;
                    break;
                }
                // '.'が見つかった場合
                else if (character.GetCode() == '.')
                {
                    // 最初に見つかったのであれば少数点数と仮定する
                    if (token.type == TokenType::IntValue)
                    {
                        token.type = TokenType::FloatValue;
                    }
                    // 2回目以降の発見であればそれは文字列として扱う
                    else if (token.type == TokenType::FloatValue)
                    {
                        token.type = TokenType::String;
                    }
                }
                // 上記以外で数値とスペース以外の文字が見つかったら文字列
                else if (!character.IsSpace() && !character.IsDigit())
                {
                    token.type = TokenType::String;
                }

                token.name += character.ToString();
            }

            // 文字列の場合は末尾のスペースを削除する
            if (token.type == TokenType::String)
            {
                for (;(token.name.back() == ' ') || (token.name.back() == '\t');)
                {
                    token.name.pop_back();
                }
            }
        }
        // '='はEqual
        else if (line[index].GetCode() == '=')
        {
            token.name = "=";
            token.type = TokenType::Equal;
        }
        // それ以外は文字列
        else
        {
            // '='が見つかるか終端に達するまで走査
            token.type = TokenType::String;
            for (uint32_t i = 0;; i++)
            {
                const auto &character = line[index + i];

                if (character.IsNull() || character.GetCode() == '=')
                {
                    nextIndex = index + i;
                    break;
                }

                token.name += character.ToString();
            }

            // 末尾のスペースを削除
            for (;(token.name.back() == ' ') || (token.name.back() == '\t');)
            {
                token.name.pop_back();
            }
        }

        // 数値を文字列から値に変換
        if (token.type == TokenType::IntValue)
        {
            token.intValue = std::stoi(token.name);
        }
        else if (token.type == TokenType::FloatValue)
        {
            token.floatValue = std::stof(token.name);
        }

        // 文字列の内容が特定のものであれば論理値として扱う
        if (!isForcedString && (token.type == TokenType::String))
        {
            std::string name = token.name;
            for (auto &c : name)
            {
                c = static_cast<char>(std::tolower(c));
            }
            if ((name == "true") || (name == "yes"))
            {
                token.type = TokenType::IntValue;
                token.intValue = 1;
            }
            else if ((name == "false") || (name == "no"))
            {
                token.type = TokenType::IntValue;
                token.intValue = 0;
            }
        }

        // トークンを追加して次の文字列へ
        tokenArray.push_back(token);
        index = nextIndex;
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      トークンを解析
 *  \param[in]  tokenArray  トークン配列
 *  \param[in]  section     現在のセクション
 *  \retval     true        成功
 *  \retval     false       失敗（構文エラー）
 */
/* ------------------------------------------------------------------------- */
bool ConfigParser::ParseToken(const TokenArray &tokenArray, std::string section) noexcept
{
    size_t arraySize = tokenArray.size();

    // NOTE: 仕様が単純なので真面目にマッチング判定を行わない

    // トークンが1つ
    if (arraySize == 1)
    {
        // セクション名であれば現在のセクション名を変更
        if (tokenArray[0].type == TokenType::SectionName)
        {
            _currentSection = tokenArray[0].name;
        }
        // それ以外は構文エラー
        else
        {
            return false;
        }
    }
    // トークンが2つ
    else if (arraySize == 2)
    {
        // "key ="の状態は意図的な省略とみなしてその記述を無視する
        if (IsValidKeyOrValue(tokenArray[0].type) && (tokenArray[1].type == TokenType::Equal))
        {
            return true;
        }
    }
    // トークンが3つ
    else if (arraySize == 3)
    {
        // 2つ目がEqualでなければ構文エラー
        if (tokenArray[1].type != TokenType::Equal)
        {
            return false;
        }
        // Equalの両辺がキーや値として認識可能でなければ構文エラー
        else if (!IsValidKeyOrValue(tokenArray[0].type) || !IsValidKeyOrValue(tokenArray[2].type))
        {
            return false;
        }

        AddConfigValue(section, tokenArray[0].name, tokenArray[2]);
    }
    // トークンが0なのは空行（コメント含む）なので何もしない
    else if (arraySize == 0)
    {
        return true;
    }
    // それ以上トークンが存在する場合は構文エラー
    else
    {
        return false;
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      そのトークンタイプがキーか値になれるかを取得
 *  \param[in]  tokenType   トークンの種類
 *  \retval     true        キーか値になれる
 *  \retval     false       キーにも値にもなれない
 */
/* ------------------------------------------------------------------------- */
bool ConfigParser::IsValidKeyOrValue(ConfigParser::TokenType tokenType) noexcept
{
    switch (tokenType)
    {
        case TokenType::String:
        case TokenType::IntValue:
        case TokenType::FloatValue:
            return true;

        default:
            break;
    }

    return false;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      設定値を追加
 *  \param[in]  section     セクション名
 *  \param[in]  key         キー名
 *  \param[in]  value       追加する値（トークンそのまま）
 */
/* ------------------------------------------------------------------------- */
void ConfigParser::AddConfigValue(const std::string section, const std::string key, const Token &value) noexcept
{
    std::string addKey;

    // セクション名が空ならキーそのまま
    if (section.empty())
    {
        addKey = key;
    }
    // 空でなければキーの頭にセクション名を付ける
    else
    {
        addKey = section + std::string(".") + key;
    }

    // 検索して見つからなければ新たに追加
    auto it = _valueArray.find(addKey);
    if (it == _valueArray.end())
    {
        _valueArray.emplace(addKey, value);
    }
    // 見つかれば上書き
    else
    {
        it->second = value;
    }
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      指定の文字列がコマンドラインオプションのキーになれるかを取得
 *  \retval     true    なれる
 *  \retval     false   なれない
 */
/* ------------------------------------------------------------------------- */
bool ConfigParser::IsArgsOptionKey(const char *argv) noexcept
{
    size_t length = strlen(argv);

    // 2文字以下はキーではない
    if (length <= 2)
    {
        return false;
    }

    // 先頭が"--"かをチェック
    if ((argv[0] != '-') || (argv[1] != '-'))
    {
        return false;
    }

    return true;
}

}   // namespace MTK

// vim: et ts=4 sw=4 sts=4

