// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       csv_parser.cc
 *  \brief      CSVパーサークラス
 *  \date       Since: October 17, 2018. 15:32:18 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <mtk/csv_parser.h>

namespace MTK
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      読み込み
 *  \param[in]  csvPath     CSVファイルのパス
 *  \param[in]  encoding    読み込むテキストのエンコードタイプ
 *  \retval     true        成功
 *  \retval     false       失敗
 */
/* ------------------------------------------------------------------------- */
bool CSVParser::Load(const std::filesystem::path &path, TextEncoding encoding) noexcept
{
    TextStream stream;
    if (!stream.Load(path, encoding))
    {
        return false;
    }

    return ParseCSV(stream);
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          CSVをパース
 *  \param[in,out]  stream  文字列ストリーム
 *  \retval         true    成功
 *  \retval         false   失敗
 */
/* ------------------------------------------------------------------------- */
bool CSVParser::ParseCSV(TextStream &stream) noexcept
{
    std::string csvField;

    _csvFieldArray.clear();
    _columnCount = 0;
    _recordCount = 0;

    bool isTerminatedParse = false;
    while (!isTerminatedParse)
    {
        auto character = stream.Read();
        if (!character.IsValid())
        {
            return false;
        }

        bool isTerminatedField = false;
        bool isTerminatedRecord = false;

        // '"'が来たら'"'で終了するまで読み続ける
        if (character.GetCode() == '"')
        {
            if (!ParseEnclosedField(stream, csvField, isTerminatedRecord, isTerminatedParse))
            {
                return false;
            }

            isTerminatedField = true;
        }
        // ','が来たらフィールドの終了
        else if (character.GetCode() == ',')
        {
            isTerminatedField = true;
        }
        // 改行が来たらレコードの終了
        else if (character.GetCode() == 0x0D)
        {
            // RFC4180に準拠して，レコードの終了はCR+LFでないと不正扱いにする
            character = stream.Read();
            if (!character.IsValid() || (character.GetCode() != 0x0A))
            {
                return false;
            }
            isTerminatedField = true;
            isTerminatedRecord = true;
        }
        // 終了コードが来たらレコードの終了かつ処理の終了
        // 最後のレコードは','を省略できるため，直接breakはできない
        else if (character.IsNull())
        {
            isTerminatedField = true;
            isTerminatedRecord = true;
            isTerminatedParse = true;
        }
        // それ以外はフィールドの文字を追加
        else
        {
            csvField += character.ToString();
        }

        // フィールドが終了していたら配列に登録して次の準備
        if (isTerminatedField)
        {
            if (_recordCount == 0)
            {
                _columnCount++;
            }
            if (isTerminatedRecord)
            {
                _recordCount++;
            }

            _csvFieldArray.push_back(csvField);
            csvField.clear();
        }
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief          '"'で囲まれたフィールドのパース
 *  \param[in,out]  stream              文字列ストリーム
 *  \param[out]     csvField            フィールドの文字の格納先
 *  \param[out]     isTerminatedRecord  レコードの終了
 *  \param[out]     isTerminatedParse   パースの終了
 *  \retval         true                成功
 *  \retval         false               失敗（続行不可）
 */
/* ------------------------------------------------------------------------- */
bool CSVParser::ParseEnclosedField(
    TextStream &stream,
    std::string &csvField,
    bool &isTerminatedRecord,
    bool &isTerminatedParse
) noexcept
{
    while (true)
    {
        auto character = stream.Read();

        // 不正な文字 or 末尾の'"'が見つからないまま終端に達したら不正
        if (!character.IsValid() || character.IsNull())
        {
            return false;
        }
        // '"'の検出
        else if (character.GetCode() == '"')
        {
            character = stream.Read();

            // 不正文字チェック
            if (!character.IsValid())
            {
                return false;
            }
            // '"'が連続している場合は単体の'"'として扱う
            else if (character.GetCode() == '"')
            {
                csvField += '"';
            }
            // '"'で終了した後に','が来ていたらフィールドの終了
            else if (character.GetCode() == ',')
            {
                break;
            }
            // '"'で終了した後にCRが来ていたらレコードの終了
            else if (character.GetCode() == 0x0D)
            {
                // CRの後がLFでなければ不正扱い
                character = stream.Read();
                if (!character.IsValid() || (character.GetCode() != 0x0A))
                {
                    return false;
                }
                isTerminatedRecord = true;
                break;
            }
            // ファイルの末尾もレコードの終了扱い
            else if (character.IsNull())
            {
                isTerminatedRecord = true;
                isTerminatedParse = true;
                break;
            }
            // それ以外が来ていたら不正
            else
            {
                return false;
            }
        }
        // '"'でなければその文字を追加
        else
        {
            csvField += character.ToString();
        }
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      フィールドの取得
 *  \param[in]  record  レコードのインデックス
 *  \param[in]  column  フィールドの列
 *  \return     フィールドの値
 */
/* ------------------------------------------------------------------------- */
std::string CSVParser::GetField(int record, int column) const noexcept
{
    std::string field;
    if ((record < 0) || (record >= _recordCount))
    {
        return field;
    }
    else if ((column < 0) || (column >= _columnCount))
    {
        return field;
    }

    size_t index = static_cast<size_t>(record) * static_cast<size_t>(_columnCount) + static_cast<size_t>(column);
    if (index >= _csvFieldArray.size())
    {
        return "";
    }
    return _csvFieldArray[index];
}
}   // namespace MTK

// vim: et ts=4 sw=4 sts=4
