// SPDX-License-Identifier: Zlib
/* ------------------------------------------------------------------------- */
/*!
 *  \file       text_character.cc
 *  \brief      文字クラス
 *  \date       Since: October 4, 2018. 2:20:42 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <mtk/text_character.h>
#include <string.h>

namespace MTK
{
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      コードをバイト列に変換
 *  \param[out] byteArray   バイトコードの格納先
 *  \retval     true        成功
 *  \retval     false       失敗
 *  \note
 *      とりあえずUTF-8で返す．
 *      将来は他の他の文字コード（Shift-JISなど）にも対応するかも．
 */
/* ------------------------------------------------------------------------- */
bool Character::ToByte(char byteArray[kByteArraySize]) const noexcept
{
    if (!IsValid())
    {
        return false;
    }

    memset(byteArray, 0, kByteArraySize);

    // 1Byte
    if (_code <= 0x7F)
    {
        byteArray[0] = static_cast<char>(_code);
    }
    // 2Byte
    else if (_code <= 0x7FF)
    {
        uint32_t code = _code & 0x7FF;
        byteArray[0] = static_cast<char>(0xC0 | (code >> 6));
        byteArray[1] = static_cast<char>(0x80 | (code & 0x3F));
    }
    // 3Byte
    else if (_code <= 0xFFFF)
    {
        uint32_t code = _code & 0xFFFF;
        byteArray[0] = static_cast<char>(0xE0 | (code >> 12));
        byteArray[1] = static_cast<char>(0x80 | ((code >> 6) & 0x3F));
        byteArray[2] = static_cast<char>(0x80 | (code & 0x3F));
    }
    // 4Byte
    else if (_code <= 0x1FFFFF)
    {
        uint32_t code = _code & 0x1FFFFF;
        byteArray[0] = static_cast<char>(0xF0 | (code >> 18));
        byteArray[1] = static_cast<char>(0x80 | ((code >> 12) & 0x3F));
        byteArray[2] = static_cast<char>(0x80 | ((code >> 6) & 0x3F));
        byteArray[3] = static_cast<char>(0x80 | (code & 0x3F));
    }
    // それ以降はUTF-8では不正なシーケンスなので失敗
    else
    {
        return false;
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      文字をstd::stringに変換
 *  \return     変換した文字
 */
/* ------------------------------------------------------------------------- */
std::string Character::ToString() const noexcept
{
    char byteArray[kByteArraySize];
    if (ToByte(byteArray))
    {
        return std::string(byteArray);
    }

    return std::string("");
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      文字列をstd::stringに変換
 *  \return     変換した文字列
 */
/* ------------------------------------------------------------------------- */
std::string CharacterArray::ToString() const
{
    std::string str("");

    for (const auto &character : *this)
    {
        str += character.ToString();
    }

    return str;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      文字列の一部をstd::stringに変換
 *  \param[in]  start   開始位置
 *  \param[in]  end     終了位置
 *  \return     startからendまでの文字を変換した文字列
 */
/* ------------------------------------------------------------------------- */
std::string CharacterArray::ToString(uint32_t start, uint32_t end) const
{
    uint32_t arraySize = static_cast<uint32_t>(size());
    if ((end == 0) || (end >= arraySize))
    {
        end = arraySize - 1;
    }

    std::string str("");

    for (uint32_t i = start; i < end; i++)
    {
        str += at(i).ToString();
    }

    return str;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      文字検索
 *  \param[in]  findCode    検索する文字（Unicodeスカラ値）
 *  \param[in]  start       検索開始位置
 *  \return     見つかった場合はそのインデックス，見つからない場合は負数．
 */
/* ------------------------------------------------------------------------- */
int CharacterArray::Find(uint32_t findCode, uint32_t start) const
{
    uint32_t arraySize = static_cast<uint32_t>(size());

    for (uint32_t i = start; i < arraySize; i++)
    {
        if (at(i).GetCode() == findCode)
        {
            return static_cast<int>(i);
        }
    }

    return -1;
}
}   // namespace MTK

// vim: et ts=4 sw=4 sts=4

