/* ------------------------------------------------------------------------- */
/*!
 *  \file       run_options.cc
 *  \brief      実行オプション
 *  \date       Since: March 13, 2020. 16:23:01 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include "run_options.h"

#include "message.h"

#if defined(MTK_TARGET_WINDOWS)
#include <Windows.h>
#endif

namespace texatlas
{
namespace
{
constexpr uint32_t kMarginMax = 32;
constexpr uint32_t kSizeMax = 65536;
}

/* ------------------------------------------------------------------------- */
/*!
 *  \brief      コンストラクタ
 */
/* ------------------------------------------------------------------------- */
RunOptions::RunOptions() noexcept
    : _canExecute(false)
    , _shouldPrintVersion(false)
    , _shouldPrintHelp(false)
    , _options()
{
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      コマンドライン引数をパース
 *  \param[in]  argc    コマンドライン引数の数
 *  \param[in]  argv    コマンドライン引数
 *  \retval     true    成功
 *  \retval     false   失敗（実行不可）
 */
/* ------------------------------------------------------------------------- */
bool RunOptions::ParseCommandlineArgs(int argc, const char *argv[]) noexcept
{
    // 引数が無かったら簡易ヘルプを表示
    if (argc <= 1)
    {
        _shouldPrintHelp = true;
        return true;
    }

    for (int i = 1; i < argc; i++)
    {
        const char *arg = argv[i];
        const char *nextArg = (i < (argc - 1)) ? argv[i + 1] : nullptr;

        // バージョン情報表示
        if ((strcmp(arg, "-v") == 0 ) || (strcmp(arg, "--version") == 0))
        {
            _shouldPrintVersion = true;
            return true;
        }
        // 簡易ヘルプ表示
        else if (strcmp(arg, "--help") == 0)
        {
            _shouldPrintHelp = true;
            return true;
        }
        // 出力ディレクトリ
        else if ((strcmp(arg, "-o") == 0 ) || (strcmp(arg, "--output") == 0))
        {
            if (nextArg == nullptr)
            {
                PutError(MsgID::ParameterIsMissing, {arg});
                return false;
            }
            _options.outputPath = nextArg;
            i++;
        }
        // 出力名
        else if ((strcmp(arg, "-n") == 0 ) || (strcmp(arg, "--name") == 0))
        {
            if (nextArg == nullptr)
            {
                PutError(MsgID::ParameterIsMissing, {arg});
                return false;
            }
            _options.outputName = nextArg;
            i++;
        }
        // マージン
        else if ((strcmp(arg, "-m") == 0) || (strcmp(arg, "--margin") == 0))
        {
            if (nextArg == nullptr)
            {
                PutError(MsgID::ParameterIsMissing, {arg});
                return false;
            }
            if (!ToValue(_options.margin, nextArg))
            {
                PutError(MsgID::IncorrectMarginParameter, {nextArg});
                return false;
            }
            i++;
        }
        // サイズ
        else if ((strcmp(arg, "-s") == 0) || (strcmp(arg, "--size") == 0))
        {
            if (nextArg == nullptr)
            {
                PutError(MsgID::ParameterIsMissing, {arg});
                return false;
            }
            if (!ParseSize(_options.width, _options.height, nextArg))
            {
                PutError(MsgID::IncorrectSizeParameter, {nextArg});
                return false;
            }
            i++;
        }
        // ハッシュシード
        else if (strcmp(arg, "--hash-seed") == 0)
        {
            if (nextArg == nullptr)
            {
                PutError(MsgID::ParameterIsMissing, {arg});
                return false;
            }
            if (!ToValue(_options.hashSeed, nextArg))
            {
                PutError(MsgID::IncorrectHashParameter, {nextArg});
                return false;
            }
            i++;
        }
        // 冗舌モード
        else if (strcmp(arg, "--verbose") == 0)
        {
            _options.isVerbose = true;
        }
        // 不明なオプションは飛ばす
        else if (arg[0] == '-')
        {
        }
        // 上記以外は対象ファイル
        else
        {
#if defined(MTK_TARGET_WINDOWS)
            auto paths = ExpandingWildcard(arg);
            for (const auto &path : paths)
            {
                _options.targetFiles.push_back(path);
            }
#else
            _options.targetFiles.push_back(arg);
#endif
        }
    }

    // 対象ファイルチェック
    if (_options.targetFiles.empty())
    {
        PutError(MsgID::TargetNotSpecified);
        return false;
    }

    // サイズチェック
    if ((_options.width == 0) || (_options.width > kSizeMax) ||
        (_options.height == 0) || (_options.height > kSizeMax))
    {
        PutError(MsgID::OutOfRangeOutputSize, {kSizeMax});
        return false;
    }

    // マージンチェック
    if (_options.margin > kMarginMax)
    {
        PutError(MsgID::OutOfRangeMarginSize, {kSizeMax});
        return false;
    }

    _canExecute = true;

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      文字列から値を取得
 *  \param[out] dest    出力先
 *  \param[in]  arg     引数の文字列
 *  \retval     true    成功
 *  \retval     false   失敗（引数が不正）
 */
/* ------------------------------------------------------------------------- */
bool RunOptions::ToValue(uint32_t &dest, const std::string &arg) const noexcept
{
    try
    {
        auto i = std::stoi(arg);
        if (i < 0)
        {
            return false;
        }
        else
        {
            dest = static_cast<uint32_t>(i);
        }
    }
    catch ([[maybe_unused]] std::exception e)
    {
        return false;
    }

    return true;
}


/* ------------------------------------------------------------------------- */
/*!
 *  \brief      サイズの文字列をパース
 *  \param[out] width   幅の出力先
 *  \param[out] height  高さの出力先
 *  \param[in]  arg     引数の文字列
 *  \retval     true    成功
 *  \retval     false   失敗（引数が不正）
 */
/* ------------------------------------------------------------------------- */
bool RunOptions::ParseSize(uint32_t &width, uint32_t &height, const std::string &arg) const noexcept
{
    auto pos = arg.find('x');
    if (pos == std::string::npos)
    {
        return false;
    }

    if (!ToValue(width, arg.substr(0, pos)))
    {
        return false;
    }

    if (!ToValue(height, arg.substr(pos + 1)))
    {
        return false;
    }

    return true;
}


#if defined(MTK_TARGET_WINDOWS)
/* ------------------------------------------------------------------------- */
/*!
 *  \brief      Windowsのワイルドカード展開
 *  \param[in]  arg     展開する引数
 *  \return     展開後のパス
 */
/* ------------------------------------------------------------------------- */
std::vector<std::filesystem::path> RunOptions::ExpandingWildcard(const char *arg) const noexcept
{
    std::vector<std::filesystem::path> paths;
    std::filesystem::path basePath(arg);
    WIN32_FIND_DATA findFileData;

    auto handle = FindFirstFile(basePath.c_str(), &findFileData);
    if (handle == INVALID_HANDLE_VALUE)
    {
        return paths;
    }

    do
    {
        auto path = basePath.parent_path();
        path /= findFileData.cFileName;
        paths.push_back(path);
    } while (FindNextFile(handle, &findFileData));

    return paths;
}
#endif

}   // namespace texatlas

// vim: et ts=4 sw=4 sts=4
