/* ------------------------------------------------------------------------- */
/*!
 *  \file       run_options.h
 *  \brief      実行オプション
 *  \date       Since: March 13, 2020. 16:23:02 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#ifndef INCGUARD_RUN_OPTIONS_H_1584084182
#define INCGUARD_RUN_OPTIONS_H_1584084182

#include <filesystem>
#include <string>
#include <vector>

#include <mtk/text_hash.h>
#include <mtk/environment.h>

namespace texatlas
{
//! 実行オプションクラス
class RunOptions
{
public:
    RunOptions() noexcept;

    //! 実行オプション値
    struct Options
    {
        std::filesystem::path outputPath;     //!< ファイルの出力パス
        std::filesystem::path outputName;     //!< 出力ファイル名
        std::vector<std::filesystem::path> targetFiles;
        bool isVerbose;
        uint32_t margin;            
        uint32_t width;
        uint32_t height;
        uint32_t hashSeed;

        //! 初期値
        Options() noexcept
            : outputPath("./")
            , outputName("texture")
            , targetFiles()
            , isVerbose(false)
            , margin(1)
            , width(1024)
            , height(1024)
            , hashSeed(MTK::Hash::kFNV1aDefaultValue32)
        {}
    };

    bool ParseCommandlineArgs(int argc, const char *argv[]) noexcept;

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      実行可能かどうかを取得
     *  \retval     true    実行可能
     *  \retval     false   実行不可能
     */
    /* ------------------------------------------------------------------------- */
    constexpr bool CanExecute() const noexcept
    {
        return _canExecute;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      バージョン情報の表示かどうかを取得
     *  \retval     true    バージョン情報を表示する
     *  \retval     false   バージョン情報を表示しない
     */
    /* ------------------------------------------------------------------------- */
    constexpr bool ShouldPrintVersion() const noexcept
    {
        return _shouldPrintVersion;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      簡易ヘルプの表示かどうかを取得
     *  \retval     true    簡易ヘルプを表示する
     *  \retval     false   簡易ヘルプを表示しない
     */
    /* ------------------------------------------------------------------------- */
    constexpr bool ShouldPrintHelp() const noexcept
    {
        return _shouldPrintHelp;
    }

    /* ------------------------------------------------------------------------- */
    /*!
     *  \brief      オプションの取得
     *  \return     オプション情報
     */
    /* ------------------------------------------------------------------------- */
    constexpr const Options &GetOptions() noexcept
    {
        return _options;
    }

private:
    bool ParseSize(uint32_t &width, uint32_t &height, const std::string &arg) const noexcept;
    bool ToValue(uint32_t &dest, const std::string &arg) const noexcept;

#if defined(MTK_TARGET_WINDOWS)
    std::vector<std::filesystem::path> ExpandingWildcard(const char *arg) const noexcept;
#endif

    bool _canExecute;               //!< 実行可能フラグ
    bool _shouldPrintVersion;       //!< バージョン情報表示フラグ
    bool _shouldPrintHelp;          //!< 簡易ヘルプ表示フラグ

    Options _options;       //!< 実行オプション
};
}   // namespace texatlas

#endif  // INCGUARD_RUN_OPTIONS_H_1584084182

// vim: et ts=4 sw=4 sts=4
