/* ------------------------------------------------------------------------- */
/*!
 *  \file       texatlas.cc
 *  \brief      テクスチャのアトラス化ツール
 *  \date       Since: June 6, 2021. 16:55:39 JST.
 *  \author     Acerola
 */
/* ------------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>

#include "atlas_generator.h"
#include "message.h"
#include "run_options.h"

constexpr const char *kProgramName = u8"mgl-texatlas";  // プログラム名
constexpr const char *kVersionInfo = u8"1.3.0";         // バージョン

/* ------------------------------------------------------------------------- */
/*!
 *  \brief      バージョン情報の表示
 */
/* ------------------------------------------------------------------------- */
void PrintVersionInfo() noexcept
{
    printf("%s version %s (compiled at %s %s)\n", kProgramName, kVersionInfo, __DATE__, __TIME__);
}

/* ------------------------------------------------------------------------- */
/*!
 *  \brief      メイン
 *  \param[in]  argc    コマンドライン引数の数
 *  \param[in]  argv    コマンドライン引数
 *  \return     実行結果
 */
/* ------------------------------------------------------------------------- */
int main(int argc, const char *argv[])
{
    using namespace texatlas;

    InitMessage();

    RunOptions runOptions;
    runOptions.ParseCommandlineArgs(argc, argv);

    // バージョンの表示
    if (runOptions.ShouldPrintVersion())
    {
        PrintVersionInfo();
        return EXIT_SUCCESS;
    }
    // 簡易ヘルプの表示
    else if (runOptions.ShouldPrintHelp())
    {
        PrintVersionInfo();
        PutMessage(MsgID::HelpText);
        return EXIT_SUCCESS;
    }
    // 実行不可能であればすぐ抜ける
    else if (!runOptions.CanExecute())
    {
        return EXIT_FAILURE;
    }

    auto option = runOptions.GetOptions();
    AtlasGenerator generator(option.isVerbose);

    PutMessage(MsgID::UseHashSeedValue, {option.hashSeed});

    // 入力された対象ファイルからリストを生成
    auto targetCount = generator.CreateTargetList(option.targetFiles, option.hashSeed);
    if (targetCount == 0)
    {
        PutError(MsgID::TargetNotFound);
        return EXIT_FAILURE;
    }
    if (option.isVerbose)
    {
        generator.PrintTargetList();
    }

    // ハッシュが衝突していないかチェック
    if (generator.CheckCollidingHash() != 0)
    {
        return EXIT_FAILURE;
    }

    // 生成を実行
    if (!generator.Generate(option.width, option.height, option.margin, option.outputName.u8string().c_str()))
    {
        return EXIT_FAILURE;
    }

    // 結果を保存
    if (!generator.Save(option.outputPath, option.outputName))
    {
        return EXIT_FAILURE;
    }

    PutMessage(MsgID::Succeeded);

    return EXIT_SUCCESS;
}

// vim: et ts=4 sw=4 sts=4
